#ifndef Layout_hh
#define Layout_hh


#include "Array.hh"
#include "Graph.hh"
#include <LEDA/string.h>


/**
	@short Data structure to handle linear graph layouts.
	@author Jordi Petit i Silvestre <jpetit@lsi.upc.es>

	Graph is a data structure to handle linear graph layouts.
	
	This a relatively low level class, in the sense that many of its
	invariants must be maintened "by hand" in order to have a better
	efficiency.
*/


class Layout: public Array<int>
{
	private:
	
		/** Binded graph */
		const Graph* _G;		
		
	public:
		
		/** Constructor: creates a canonical Layout for a binded graph */
		Layout (const Graph& G);
		

		/** Returns the binded graph. */
		inline const Graph& graph () const {return *_G;}
		/** Returns the number of nodes. */
		inline int nodes () const {return size();}
		/** Returns the number of edges of the binded graph. */
		inline int edges () const {return _G->edges();}
		/** Returns the position of node u. This the the same than using operator[]. */
		inline const int& phi (const int u) const {return operator[](u);}	
		/** Returns a reference to the position of node u. This the the same than using operator[]. */
		inline int& phi (const int u) {return operator[](u);}


		/** Input in "lay" format */
		void get_lay (istream& is);
		/** Input in "lay" format */
		void get_lay (const string& filename);
		/** Output in "lay" format */
		void put_lay (ostream& os) const;
		/** Output in "lay" format */
		void put_lay (const string& filename) const;


		/** Set to the canonical layout */
		void set_canonical ();
		/** Set to the spectral layout */
		void set_spectral ();
		/** Set to a random layout */
		void set_random ();


		/** Returns the bisection cost.  */
		int bi () const;
		/** Returns the bandwidth cost.  */
		int bw () const;
		/** Returns the cutwidth cost.  */
		int cw () const;
		/** Returns the linear arrangement cost.  */
		int la () const;
		/** Returns the sum cut cost.  */
		int sc () const;
		/** Returns the vertex separation cost.  */
		int vs () const;
		/** Returns all the costs. */
		void all (int& bi, int& bw, int& cw, int& la, int& sc, int& vs) const;
		/** Returns all the costs as a string (swig). */
		string all () const;
		/** Returns the edge length of the edge between 'u' and 'v'. */
		inline int length (const int u, const int v) const {return abs(phi(u)-phi(v));}


		/** Check if the layout is consistent */
		bool check (bool abort=true) const;

		/** Swaps the position of two nodes 'u' and 'v' in the layout. */
		inline void flip2 (const int u, const int v) {
			swap(u,v);
		}
		
		/** Rotates to the left the position of three nodes 'u', 'v' and 'w' in the layout. */
		inline void flip3 (const int u, const int v, const int w) {
			const int tmp=phi(u); phi(u)=phi(v); phi(v)=phi(w); phi(w)=tmp;
		}

		/** Rotates to the right the position of three nodes 'u', 'v' and 'w' in the layout. */
		inline void unflip3 (const int u, const int v, const int w) {
			const int tmp=phi(w); phi(w)=phi(v); phi(v)=phi(u); phi(u)=tmp;
		}


		/** Computes the variation of the LA cost when flipping two nodes. */
		inline int la_delta2 (const int u, const int v) {
			const Graph& G=*_G;
			if (u==v) return 0;
			int d=0;
			for (int i=G.adj(u); i<G.adj(u+1); i++) d-=length(u,G.edg(i));
			for (int i=G.adj(v); i<G.adj(v+1); i++) d-=length(v,G.edg(i));
			flip2(u,v);
			for (int i=G.adj(u); i<G.adj(u+1); i++) d+=length(u,G.edg(i));
			for (int i=G.adj(v); i<G.adj(v+1); i++) d+=length(v,G.edg(i));
			flip2(u,v);
			return d;
		}

		/** Computes the variation of the LA cost when flipping three nodes. */
		inline int la_delta3 (const int u, const int v, const int w) {
			const Graph& G=*_G;
			if (u==v && v==w) return 0;
			if (u==v) return la_delta2(v,w);
			if (u==w) return la_delta2(u,w);
			if (v==w) return la_delta2(u,v);
			int d=0;
			for (int i=G.adj(u); i<G.adj(u+1); i++) d-=length(u,G.edg(i));
			for (int i=G.adj(v); i<G.adj(v+1); i++)	if (G.edg(i)!=u) d-=length(v,G.edg(i));
			for (int i=G.adj(w); i<G.adj(w+1); i++)	if (G.edg(i)!=v && G.edg(i)!=u) d-=length(w,G.edg(i));	
			flip3(u,v,w);	
			for (int i=G.adj(u); i<G.adj(u+1); i++) d+=length(u,G.edg(i));			
			for (int i=G.adj(v); i<G.adj(v+1); i++)	if (G.edg(i)!=u) d+=length(v,G.edg(i));				
			for (int i=G.adj(w); i<G.adj(w+1); i++)	if (G.edg(i)!=v && G.edg(i)!=u)	d+=length(w,G.edg(i));	
			unflip3(u,v,w);
			return d;
		}

		

};


#endif

