#include "Minla.hh"


#include <assert.h>
#include "Random.hh"



// ****************************************************************************
// Utilities
// ****************************************************************************


template <class T>
inline T sqr (const T x) {
	return x*x;
}


inline double log2 (const double x) {
	return log(x)/log(2.0);
}


// ****************************************************************************
// Canonical
// ****************************************************************************


int Minla::canonical (const Graph& G, Layout& L)
{
	L=Layout(G);
	L.set_canonical();
	L.check();
	return L.la();
}


// ****************************************************************************
// Spectral
// ****************************************************************************


int Minla::spectral (const Graph& G, Layout& L)
{
	L=Layout(G);
	L.set_spectral();
	L.check();
	return L.la();
}


// ****************************************************************************
// Random
// ****************************************************************************


int Minla::random (const Graph& G, Layout& L)
{
	L=Layout(G);
	L.set_random();
	L.check();
	return L.la();
}


// ****************************************************************************
// Optimal
// ****************************************************************************


class BranchAndBound
{
	public:

		BranchAndBound (const Graph& G, const int ub=-1);
		BranchAndBound& operator() ();
		inline const Graph& graph () const {return *_G;} 
		inline const Layout& solution () const {return _L;}

	private:
		const Graph* _G;
		Layout _L;
		Layout _W;
		int _ub;
		const int _n;
		Array<bool> _avail;
		Array<int> _plb;

		void BnN (const int c, const int u);
};





BranchAndBound::BranchAndBound (const Graph& G, const int ub)
:	_G(&G),
	_L(G),
	_W(G),
	_ub(ub),
	_n(G.nodes()),
	_avail(G.nodes()),
	_plb(G.nodes()+1)
{
}


BranchAndBound& BranchAndBound::operator() ()
{
	// Init _avail
	for (int u=0; u<_n; u++) {
		_avail[u]=true;
	}

	// Compute partial lower bounds
	for (int u=0; u<_n; u++) {
		const int d=graph().deg(u);
		_plb[u]=((d%2==0) ? d*d/4+d/2 : (d*d+2*d+1)/4)/2;
	}	
	for (int u=_n-2; u>=0; u--) {
		_plb[u]+=_plb[u+1];
	}
	_plb[_n]=0;

	if (_ub<0) {
		_L.set_spectral();
		_ub=_L.la();
	}
	// Recurse!
	BnN(0,0);	
	
	return *this;
}




void BranchAndBound::BnN (const int c, const int u) 
{
	if (u==graph().nodes()) {	
		if (c<_ub) {
			_ub=c;
			_L=_W;
		}
	} else {
		for (int i=0; i<graph().nodes(); i++) {
			if (_avail[i]) {
				int c2=c;
				_avail[i]=false;
				_W[u]=i;
				for (int j=graph().adj(u); j<graph().adj(u+1); j++) {
					const int v=graph().edg(j);
					if (v<u) {
						c2+=_W.length(u,v);
				}	}
				if (c2+_plb[u+1]<_ub) {
					BnN(c2,u+1);					
				}
				_avail[i]=true;
}	}	}	}
	



int Minla::optimal (const Graph& G, Layout& L)
{
	BranchAndBound BB(G);
	BB();
	L=BB.solution();
	L.check();
	return L.la();
}




// ****************************************************************************
// Greedy
// ****************************************************************************



class Greedy
{
	public:

		Greedy (const Graph& G);
		virtual ~Greedy();

		Greedy& operator() ();
		inline const Layout& solution () const {return _L;}
		inline const Graph& graph () const {return *_G;};

	protected:

		virtual void set_ordering () = 0;
		const Graph* _G;	
		const int _n;
		Array<int> _ordering;

	private:	

		Layout _L;
};




Greedy::Greedy (const Graph& G)
:	_G(&G),
	_n(G.nodes()),
	_ordering(G.nodes()),
	_L(G)
{
}


Greedy::~Greedy ()
{
}


Greedy& Greedy::operator() ()
{
	int i,j;
	int l;				// Leftmost position
	int r;				// Rightmost position
	int	cl;				// Cost of increment at left
	int cr;				// Cost of increment at right
	int u;				// Node to place
	int v;				// Possible neighbours of u

	set_ordering();		// Call the virtual method	
	l=-1; r=1; _L[_ordering[0]]=0;
	for (i=1; i<_n; i++) {
		u=_ordering[i];
		/* Left */
		_L[u]=l; cl=0;
		for (j=0; j<i; j++) {
			v=_ordering[j];
			if (graph().connected(u,v)) {	
				cl+=_L.length(u,v);
		}	}
		/* Right */
		_L[u]=r; cr=0;
		for (j=0; j<i; j++) {
			v=_ordering[j];
			if (graph().connected(u,v)) {
				cr+=_L.length(u,v);
		}	}
		/* Decision */
		if (cl<cr) {
			_L[u]=l; l--;
		} else {
			r++;
	}	}

	/* Normalization: put L beetween 0..n-1 */
	for (u=0; u<_n; u++) {
		_L[u]-=l+1;
	}	
	return *this;
}





class GreedyNor: public Greedy
{
	public:
		
		GreedyNor (const Graph& G): Greedy(G) {}
		
	protected:
	
		void set_ordering () 
		{
			for (int i=0; i<_n; i++) {
				_ordering[i]=i;
		}	}
};


class GreedyRnd: public Greedy
{
	public:
	
		GreedyRnd (const Graph& G): Greedy(G) {}
		
	protected:

		void set_ordering () 
		{
			for (int i=0; i<_n; i++) {
				_ordering[i]=i;
			}
			_ordering.permute();
		}
};



class GreedyBFS: public Greedy
{
	public:
	
		GreedyBFS (const Graph& G): Greedy(G) {}
		
	protected:
	
		void set_ordering () 
		{
			Array<bool> visited(_n);
			for (int i=0; i<_n; i++) visited[i]=false;
			_ordering[0]=Random(_n);
			visited[_ordering[0]]=true; 
			int i=0;
			int j=1;
			while (j<_n) {
				for (int uu=graph().adj(i); uu<graph().adj(i+1); uu++) {
					const int u=graph().edg(uu);
					if (!visited[u]) {
						_ordering[j]=u; 
						j++; 
						visited[u]=true;
				}	}
				i++;
		}	}

};


class GreedyRBS: public Greedy
{
	public:
	
		GreedyRBS (const Graph& G): Greedy(G) {}
		
	protected:
	
		void set_ordering ()
		{
			Array<int> order(_n);
			Array<int> t(_n);
			int v=Random(_n);
			for (int j=0; j<_n; j++) {
				t[j]=0;
			}
			int c=0; 
			int p=0; 
			while (c!=_n) {
				p++;
				t[v]++; 
				if (t[v]==1) {
					_ordering[c++]=v;
				}
				v=graph().edg(Random(graph().adj(v),graph().adj(v+1)-1));
		}	}

};








class GreedyDFS: public Greedy
{
	public:
	
		GreedyDFS (const Graph& G): Greedy(G) {}
		
	protected:
	
		void set_ordering ()
		{
			assert(graph().components()==1);
			for (int j=0; j<_n; j++) {
				_ordering[j]=-1;
			}
			_ctr=0;
			int v=Random(_n);
			_ordering[v]=_ctr++;
			recurse(v);
		}

		
	private:
	
		int _ctr;
		
		void recurse (const int v)
		{
			for (int uu=graph().adj(v); uu<graph().adj(v+1); uu++) {
				const int u=graph().edg(uu);
				if (_ordering[u]==-1) {
					_ordering[u]=_ctr++; 
					recurse(u);
		}	}	}
};



int Minla::greedy_nor (const Graph& G, Layout& L)
{
	GreedyNor Alg(G);
	Alg();
	L=Alg.solution();	
	L.check();
	return L.la();
}



int Minla::greedy_rnd (const Graph& G, Layout& L)
{
	GreedyRnd Alg(G);
	Alg();
	L=Alg.solution();	
	L.check();
	return L.la();
}



int Minla::greedy_rbs (const Graph& G, Layout& L)
{
	GreedyRBS Alg(G);
	Alg();
	L=Alg.solution();	
	L.check();
	return L.la();
}


int Minla::greedy_bfs (const Graph& G, Layout& L)
{
	GreedyBFS Alg(G);
	Alg();
	L=Alg.solution();	
	L.check();
	return L.la();
}


int Minla::greedy_dfs (const Graph& G, Layout& L)
{
	GreedyDFS Alg(G);
	Alg();
	L=Alg.solution();	
	L.check();
	return L.la();
}




// ****************************************************************************
// Metropolis
// ****************************************************************************



int Minla::metropolis_E (const Graph& G, Layout& L, double t)
{
	int c=L.la();					// Current cost
	const int limit=int(sqr((G.nodes())));		// Maximimal number of consecutive failures
	int f=0;						// Consecutive failures
	
	while (f<limit) {
		const int u=G.random_node();
		const int v=G.random_neighbor(u);
		const int d=L.la_delta2(u,v);
		if (d<0 || t>0 && Random(exp(-d/t))) {
			L.flip2(u,v);
			c=c+d;
			if (d<0) f=0; else f++;
		} else {
			f++;
	}	}	 
	
	L.check();
	return L.la();
}




int Minla::metropolis_2 (const Graph& G, Layout& L, double t)
{
	int c=L.la();					// Current cost
	const int limit=int(G.nodes()*log2(G.nodes()));		// Maximimal number of consecutive failures
	int f=0;						// Consecutive failures
	
	while (f<limit) {
		const int u=G.random_node();
		const int v=G.random_node();
		const int d=L.la_delta2(u,v);
		if (d<0 || t>0 && Random(exp(-d/t))) {
			L.flip2(u,v);
			c+=d;
			if (d<0) f=0; else f++;			
		} else {
			f++;
	}	}		
	
	L.check();
	return L.la();
}



int Minla::metropolis_3 (const Graph& G, Layout& L, double t)
{
	int c=L.la();					// Current cost
	const int limit=int(G.nodes()*log2(G.nodes()));		// Maximimal number of consecutive failures
	int f=0;						// Consecutive failures
	
	while (f<limit) {
		const int u=G.random_node();
		const int v=G.random_node();
		const int w=G.random_node();
		const int d=L.la_delta3(u,v,w);
		if (d<0 || t>0 && Random(exp(-d/t))) {
			L.flip3(u,v,w);
			c=c+d;
			if (d<0) f=0; else f++;
		} else {
			f++;
	}	}
	
	L.check();
	return L.la();
}





// ****************************************************************************
// Spectral Sequencing + Simulated Annealing
// ****************************************************************************


inline double random_normal_centered () 
{
	return 
		6.0
		+Random()+Random()+Random()+Random()+Random()+Random()+
		+Random()+Random()+Random()+Random()+Random()+Random();
}


inline double random_normal (const double mu, const double sigma) 
{
	return (mu + sigma*random_normal_centered());
}




void sssa_scan (const Graph& G, Layout& L, const double t, double& wAvg, double& wDev) 
{
	const int n=G.nodes(); 	// Number of nodes
	double wSum=0;			// Sum of absolute differences
	double wSum2=0;			// Sum of squares of differences
	int changes=0;			// Number of changes performed

	for (int i=int(n*sqrt(n)); i>0; i--) {
		const int u=G.random_node();
		const int v=G.random_node();
		const int d=L.la_delta2(u,v);
		if (d<0 || t>0 && Random(exp(-d/t))) {
			changes++;
			wSum+=abs(L[u]-L[v]);
			wSum2+=(L[u]-L[v])*(L[u]-L[v]);
	}	}
	wAvg=wSum/changes;
	wDev=sqrt((wSum2-sqr(wSum)/changes)/(changes-1));
}




int Minla::sssa (const Graph& G, Layout& L, const double alpha, const int beta, const double t0, const double rho)
{
 	// Get the number of nodes
 	const int n=G.nodes();
	
	// Set initial solution with spectral sequencing
	L.set_spectral();
	int c=L.la();
	
	// Create the inverted layout
	Array<int> inv(n); 
	for (int i=0; i<n; i++) inv[L[i]]=i;

	// Perform the scanning phase
	double mu,sigma;
	sssa_scan(G,L,t0,mu,sigma);
		
	double t=t0;
	const int r=int(beta*mu*n);
	do {
		const double c0=c;
		for (int i=0; i<r; i++) {
			if (i%100000==0) cerr << c << "\t" << t << endl;
			
			int len;
			do {
				len=abs(int(random_normal(mu,sigma)));
			} while  (len==0 || len>=n);			
			const int fu=Random(n-len);
			const int fv=fu+len;
			const int u=inv[fu];
			const int v=inv[fv];	
			const int d=L.la_delta2(u,v);			
			if (d<0 || t>0 && Random(exp(-d/t))) {
				L.flip2(u,v);
				inv.swap(fu,fv);
				c+=d;
			}		
		}
		if (c/c0) {
			t=alpha*t;
		}
	} while (t>0.1);
	
	L.check();
	return L.la();	
}
