%module layouts


%{
#include <LEDA/string.h>
#include "Graph.hh"
#include "Layout.hh"
#include "Minla.hh"
#include "LowerBounds.hh"
%}


%typemap(tcl,in) const string& {
	$target=new string($source);	
}


%typemap(tcl,freearg) const string& {
	delete $source;	
}


%typemap(tcl,out) string {	
	sprintf($target,"%s",~*$source);
	delete $source;
}


#define ll_version "1.0"


class Graph
{
	public:
		/** Constructor: reads the graph from a file. If filename="" creates an empty graph. */
		Graph (const string& filename);
		
		/** Input in "gra" format */
		void get_gra (const string& filename);	
		/** Output in "gra" format */
		void put_gra (const string& filename) const;
		/** Output in "cha" format */
		void put_cha (const string& filename) const;
		/** Output in "rmf" format */
		void put_rmf (const string& filename) const;
		
		int nodes () const;
		/** Returns the number of edges in the graph. */
		int edges () const;
	
		/** Answer wether u and v are connected by an edge */
		bool connected (const int u, const int v) const;
		/** Returns the maximum degree of the graph. */
		int max_degree () const;
		/** Returns the number of connected components in the graph. */
		int components () const;
		/** Returns a node from the graph selected uniformly at random. */
		int random_node () const;
		/** Returns a a random neighbor of u selected uniformly at random. */
		int random_neighbor (const int u) const;

		/** Generates a (valid)	empty graph, that is, with zero nodes and zero edges. */
		void gen_empty ();
		
		/** Generates a sample graph with 5 nodes. */
		void gen_sample ();

		/** 
			Generates a random graph in the G(n,p) model. The G(n,p) model
			is made of graphs with 'n' nodes where each pair of distinct nodes 
			is joined with an edge with probability 'p'.
		*/		
		void gen_random_np (const int n, const double p);

		/** 
			Generates a random graph in the G(n,m) model. The G(n,m) model
			is made of graphs with 'n' nodes and 'm' edges.
		*/			
		void gen_random_nm (const int n, const int m);

		/** 
			Generates a random linear graph in the RLG(n,c) model. 
			In those graphs, V=[n], and Prob{ij\in E}=c/(j-i+1) for all 1<=i<j<=n.
		*/				
		void gen_random_linear (const int n, const double c=1);	

		/**
			Generates a site percolation graph with parameter 'p'. 
			Some more exaplanation needed!!! 
		*/
		void gen_site_percolation (const double p);	

		/**
			Generates a bond percolation graph with parameter 'p'. 
			Some more exaplanation!!! Subcritical p!!!
		*/
		void gen_bond_percolation (const double p);	

		/**	Generates a complete graph with 'n' nodes. */
		void gen_complete (const int n);

		/**
			Generates a complete bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side.
		*/
		void gen_complete_bipartite (const int a, const int b);

		/**
			Generates a random bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side and 'm'
			edges.
		*/
		void gen_random_bipartite_m (const int a, const int b, const int m);

		/**
			Generates a random bipartite graph with 'a' nodes
			in one side and 'b' nodes on the other side and edge probability
			'p' for opsed side nodes.
		*/
		void gen_random_bipartite_p (const int a, const int b, const double p);

		/** Generates a square grid graph of side x. */
		void gen_grid (const int x);

};





class Layout 
{
	public:
		
		/** Constructor: creates a canonical Layout for a binded graph */
		Layout (const Graph& G);

		/** Returns the binded graph. */
		const Graph& graph () const; 
		/** Returns the number of nodes. */
		int nodes () const;
		/** Returns the number of edges of the binded graph. */
		int edges ();
		/** Returns the position of node u. This the the same than using operator[]. */
		int phi (const int u) const;

		/** Input in "lay" format */
		void get_lay (const string& filename);
		/** Output in "lay" format */
		void put_lay (const string& filename) const;

		/** Set to the canonical layout */
		void set_canonical ();
		/** Set to the spectral layout */
		void set_spectral ();
		/** Set to a random layout */
		void set_random ();

		/** Returns the bisection cost.  */
		int bi () const;
		/** Returns the bandwidth cost.  */
		int bw () const;
		/** Returns the cutwidth cost.  */
		int cw () const;
		/** Returns the linear arrangement cost.  */
		int la () const;
		/** Returns the sum cut cost.  */
		int sc () const;
		/** Returns the vertex separation cost.  */
		int vs () const;
		/** Returns the edge length of the edge between 'u' and 'v'. */
		int length (const int u, const int v) const;
		/** Returns all the costs as a string (swig). */
		string all () const;

		/** Check if the layout is consistent */
		bool check (bool abort=true) const;
};





class Minla
{		
	public:


		/** Heuristic to solve Minla using the canonical layout. 
			Modifies the layout L and returns its LA cost. */
		static int canonical (const Graph& G, Layout& L);

		/** Heuristic to solve Minla using the spectral layout. 
			Modifies the layout L and returns its LA cost. */
		static int spectral (const Graph& G, Layout& L);

		/** Heuristic to solve Minla using a random layout. 
			Modifies the layout L and returns its LA cost. */
		static int random (const Graph& G, Layout& L);

		/** Solves Minla using the optimal layout. 
			Modifies the layout L and returns its LA cost.
			Uses backtracking and is thus very slow! */
		static int optimal (const Graph& G, Layout& L);
		
		/** Solves Minla using a greedy algorithm. 
			The initial ordering of the vertices is normal.
			Modifies the layout L and returns its LA cost.*/
		static int greedy_nor (const Graph& G, Layout& L);
		
		/** Solves Minla using a greedy algorithm. 
			The initial ordering of the vertices is random.
			Modifies the layout L and returns its LA cost. */
		static int greedy_rnd (const Graph& G, Layout& L);

		/** Solves Minla using a greedy algorithm. 
			The initial ordering of the vertices is giveb
			ny a random breadth search.
			Modifies the layout L and returns its LA cost. */
		static int greedy_rbs (const Graph& G, Layout& L);

		/** Solves Minla using a greedy algorithm. 
			The initial ordering of the vertices is giveb
			ny a breadth first search starting at a random node.
			Modifies the layout L and returns its LA cost. */
		static int greedy_bfs (const Graph& G, Layout& L);

		/** Solves Minla using a greedy algorithm. 
			The initial ordering of the vertices is giveb
			ny a depth first search starting at a random node.
			Modifies the layout L and returns its LA cost. */
		static int greedy_dfs (const Graph& G, Layout& L);



		static int metropolis_2 (const Graph& G, Layout& L, double temp=-1);
		static int metropolis_3 (const Graph& G, Layout& L, double temp=-1);
		static int metropolis_E (const Graph& G, Layout& L, double temp=-1);


		static int sssa (const Graph& G, Layout& L, 
			const double alpha=0.95,
			const int beta=12,
			const double t0=10,
			const double rho=0.9999);
};



struct LowerBounds_LA {

	/** Computes a lower bound of the MINLA problem using the Gomory-Hu method based on the min-flow max-cut theorem. */
	static int gomory_hu (const Graph& G);
	/** Computes a lower bound of the MINLA problem using the Juvan-Mohar method based on spectral properties of the graph. */
	static int juvan_mohar (const Graph& G);
	/** Computes a lower bound of the MINLA problem using a method of mine. */
	static int edges_method (const Graph& G);
	/** Computes a lower bound of the MINLA problem using a method of mine. */
	static int degree_method (const Graph& G);
	/** Computes a lower bound of the MINLA problem using a method due to Juvan and Mohar. */
	static int path_method (const Graph& G);
	/** Computes a lower bound of the MINLA problem using a method of mine. */
	static int mesh_method (const Graph& G);
	
};


